@@ -1,7 +1,7 @@
 #include "encoding.h"
 // TODO the following should come from the YAML.
 #ifndef NUM_SPECD_INTCAUSES 
-#define NUM_SPECD_INTCAUSES 16
+  #define NUM_SPECD_INTCAUSES 16
 #endif
 //#define RVTEST_FIXED_LEN
 #ifndef UNROLLSZ
@@ -11,9 +11,8 @@
 //   #define rvtest_gpr_save
 // #endif
 
-
 //-----------------------------------------------------------------------
-// RV Compliance Macros
+// RV Arch Test Macros
 //-----------------------------------------------------------------------
 #ifndef RVMODEL_SET_MSW_INT
   #warning "RVMODEL_SET_MSW_INT is not defined by target. Declaring as empty macro"
@@ -68,6 +67,7 @@
 #if XLEN==64
   #define SREG sd
   #define LREG ld
+  #define LREGWU lwu
   #define REGWIDTH 8
   #define MASK 0xFFFFFFFFFFFFFFFF
 
@@ -75,6 +75,7 @@
   #if XLEN==32
     #define SREG sw
     #define LREG lw
+    #define LREGWU lw
     #define REGWIDTH 4
   #define MASK 0xFFFFFFFF
 
@@ -85,21 +86,17 @@
   #define FLREG fld
   #define FSREG fsd
   #define FREGWIDTH 8
-
+  #define SIGALIGN 8
 #else 
   #if FLEN==32
     #define FLREG flw
     #define FSREG fsw
     #define FREGWIDTH 4
-  #endif
-#endif
-
-#if XLEN==64
-  #if FLEN==32
-    #define SREG sw
-    #define LREG lW
-    #define REGWIDTH 4
-    #define MASK 0xFFFFFFFF
+    #if XLEN==64
+        #define SIGALIGN 8
+    #else
+        #define SIGALIGN 4
+    #endif
   #endif
 #endif
 
@@ -116,6 +113,36 @@
   #define CODE_REL_TVAL_MSK 0xD008 << (REGWIDTH*8-16)
 #endif
 
+#define NAN_BOXED(__val__,__width__,__max__)    \
+    .if __width__ == 32                        ;\
+        .word __val__                          ;\
+    .else                                      ;\
+        .dword __val__                         ;\
+    .endif                                     ;\
+    .if __max__ > __width__                    ;\
+        .set pref_bytes,(__max__-__width__)/32 ;\
+    .else                                      ;\
+        .set pref_bytes, 0                     ;\
+    .endif                                     ;\
+    .rept pref_bytes                           ;\
+        .word 0xffffffff                       ;\
+    .endr                                      ;
+
+
+#define ZERO_EXTEND(__val__,__width__,__max__)  \
+    .if __max__ > __width__                    ;\
+        .set pref_bytes,(__max__-__width__)/32 ;\
+    .else                                      ;\
+        .set pref_bytes, 0                     ;\
+    .endif                                     ;\
+    .rept pref_bytes                           ;\
+        .word 0                                ;\
+    .endr                                      ;\
+    .if __width__ == 32                        ;\
+        .word __val__                          ;\
+    .else                                      ;\
+        .dword __val__                         ;\
+    .endif;
 
 // ----------------------------------- CODE BEGIN w/ TRAP HANDLER START ------------------------ //
 
@@ -261,10 +288,10 @@
   /**** to a return for anything above that (which causes a mismatch)****/
   /**********************************************************************/
   mtrampoline:		// 64 or 32 entry table
-  value = 0
+  .set value, 0
   .rept NUM_SPECD_INTCAUSES     	  // located at each possible int vectors
      j	mtrap_handler + 12*(value)  //offset < +/- 1MB
-     value = value + 1
+     .set value, value + 1
   .endr
   .rept RLENG-NUM_SPECD_INTCAUSES   // fill at each impossible entry
   	mret
@@ -343,7 +370,7 @@
           bltz    t4, sv_mtval		        /* correct adjustment is code_begin in t3 */
   
           LA(     t3, mtrap_sigptr) /* adjustment assuming access is to signature region */
-          LI(t4, DATA_REL_TVAL_MSK)   /* trap#s not 14, 11..8, 2 adjust w/ data_begin */
+          LI(t4, DATA_REL_TVAL_MSK)      /* trap#s not 14, 11..8, 2 adjust w/ data_begin */
           sll     t4, t4, t2		          /* put bit# in MSB */
           bgez    t4, no_adj		          /* correct adjustment is data_begin in t3 */
   sigbound_chk:
@@ -377,8 +404,8 @@
 
   sv_mtval:
           csrr   t2, CSR_MTVAL
-          sub     t2, t2, t3		/* perform mtval adjust by either code or data position or zero*/
-          SREG      t2, 3*REGWIDTH(t1)	/* save 4th sig value, (rel mtval) into trap signature area */
+          sub    t2, t2, t3		/* perform mtval adjust by either code or data position or zero*/
+          SREG   t2, 3*REGWIDTH(t1)	/* save 4th sig value, (rel mtval) into trap signature area */
   
   resto_rtn:		/* restore and return */
           addi    t1, t1,4*REGWIDTH		/* adjust trap signature ptr (traps always save 4 words) */
@@ -556,11 +583,19 @@ rvtest_data_end:
 
 #define RVTEST_CASE(_PNAME,_DSTR,...)
 
-#define RVTEST_FP_ENABLE()\
-    LI x2, MSTATUS_FS;\
-    csrrs x3, mstatus,x0;\
-    or x2, x3, x2;\
-    csrrw x0,mstatus,x2;                      
+#define RVTEST_FP_ENABLE()              \
+ li a0, MSTATUS_FS & (MSTATUS_FS >> 1); \
+ csrs mstatus, a0;                      \
+ csrwi fcsr, 0
+
+#ifdef pext_check_vxsat_ov
+#define RVTEST_VXSAT_ENABLE()\
+ li a0, MSTATUS_VS & (MSTATUS_VS >> 1); \
+ csrs mstatus, a0;                      \
+ clrov
+#else
+#define RVTEST_VXSAT_ENABLE()
+#endif
 
 #define RVTEST_SIGBASE(_R,_TAG) \
   LA(_R,_TAG);\
@@ -573,40 +608,109 @@ rvtest_data_end:
 #define _ARG2(_1ST,_2ND, ...) _2ND
 #define _ARG1(_1ST,...) _1ST
 #define NARG(...) _ARG5(__VA_OPT__(__VA_ARGS__,)4,3,2,1,0)
-#define RVTEST_SIGUPD(_BR,_R,...)\
-  .if NARG(__VA_ARGS__) == 1;\
-    SREG _R,_ARG1(__VA_ARGS__,0)(_BR);\
-    .set offset,_ARG1(__VA_OPT__(__VA_ARGS__,)0)+REGWIDTH;\
-  .endif;\
-  .if NARG(__VA_ARGS__) == 0;\
-    SREG _R,offset(_BR);\
-  .set offset,offset+REGWIDTH;\
-  .endif;
+
+#define LOAD_MEM_VAL(_LINST, _AREG, _RD, _OFF, _TREG)  \
+    .if _OFF >= 2048                                  ;\
+        .set _off, _OFF%2048                          ;\
+        LI(_TREG, _OFF-_off)                          ;\
+        add _AREG,_AREG,_TREG                         ;\
+    .else                                             ;\
+        .set _off, _OFF                               ;\
+    .endif                                            ;\
+    _LINST _RD, _off(_AREG)                           ;\
+    .if _OFF >= 2048                                  ;\
+        sub _AREG,_AREG,_TREG                         ;\
+    .endif
+
+ /* use this function to ensure individual signature stores don't exceed offset limits */
+  /* if they would, then update the base by offset & reduce offset by -2048             */
+  /* there is an option to pre-increment offset if there was a previous signture store  */
+
+#define CHK_OFFSET(_BREG, _SZ, _PRE_INC) \
+  .if (_PRE_INC!=0)                      ;\
+    .set offset, offset+_SZ             ;\
+  .endif                                ;\
+  .if offset>=2048                      ;\
+     addi  _BREG, _BREG, (2048 - _SZ)   ;\
+     .set  offset, offset -(2048 - _SZ)  ;\
+  .endif
+
+
+ /* automatically adjust base and offset if offset gets too big */
+ /* RVTEST_SIGUPD(basereg, sigreg)        stores sigreg at offset(basereg) and updates offset by regwidth */
+ /* RVTEST_SIGUPD(basereg, sigreg,newoff) stores sigreg at newoff(basereg) and updates offset to regwidth+newoff */
+#define RVTEST_SIGUPD(_BR,_R,...)		         \
+  .if NARG(__VA_ARGS__) == 1                            ;\
+	.set offset,_ARG1(__VA_OPT__(__VA_ARGS__,0))	;\
+  .endif                                                ;\
+  CHK_OFFSET(_BR,REGWIDTH,0);\
+   SREG _R,offset(_BR)                                  ;\
+  .set offset,offset+REGWIDTH
+
+#define RVTEST_SIGUPD_F(_BR,_R,_F,...)			 \
+  .if NARG(__VA_ARGS__) == 1                            ;\
+     .set offset,_ARG1(__VA_OPT__(__VA_ARGS__,0))	;\
+  .endif                                                ;\
+  .if (offset & (SIGALIGN-1)) != 0                      ;\
+      .warning "Incorrect Offset Alignment for Signature.";\
+      .err                                              ;\
+  .endif                                                ;\
+  CHK_OFFSET(_BR,SIGALIGN,0);\
+  FSREG _R,offset(_BR)					;\
+  CHK_OFFSET(_BR,SIGALIGN,1);\
+   SREG  _F,offset(_BR)			;\
+   .set offset,offset+(SIGALIGN)
+
   
-#define RVTEST_SIGUPD_F(_BR,_R,_F,...)\
-  .if NARG(__VA_ARGS__) == 1;\
-    FSREG _R,_ARG1(__VA_ARGS__,0)(_BR);\
-    SREG _F,_ARG1(__VA_ARGS__,0)+REGWIDTH(_BR);\
-    .set offset,_ARG1(__VA_OPT__(__VA_ARGS__,)0)+(REGWIDTH+REGWIDTH);\
-  .endif;\
-  .if NARG(__VA_ARGS__) == 0;\
-    FSREG _R,offset(_BR);\
-    SREG _F,offset+REGWIDTH(_BR);\
-    .set offset,offset+(REGWIDTH+REGWIDTH);\
-  .endif;
-  
-#define RVTEST_SIGUPD_FID(_BR,_R,_F,...)\
-  .if NARG(__VA_ARGS__) == 1;\
-    SREG _R,_ARG1(__VA_ARGS__,0)(_BR);\
-    SREG _F,_ARG1(__VA_ARGS__,0)+REGWIDTH(_BR);\
-    .set offset,_ARG1(__VA_ARGS__,0)+(2*REGWIDTH);\
-  .endif;\
-  .if NARG(__VA_ARGS__) == 0;\
-    SREG _R,offset(_BR);\
-    SREG _F,offset+REGWIDTH(_BR);\
-    .set offset,offset+(2*REGWIDTH);\
-  .endif;
+#define RVTEST_SIGUPD_FID(_BR,_R,_F,...)		 \
+  .if NARG(__VA_ARGS__) == 1                            ;\
+     .set offset,_ARG1(__VA_OPT__(__VA_ARGS__,0))	;\
+  .endif                                                ;\
+  .if (offset & (SIGALIGN-1)) != 0                      ;\
+      .warning "Incorrect Offset Alignment for Signature.";\
+      .err                                              ;\
+  .endif                                                ;\
+  CHK_OFFSET(_BR,SIGALIGN,0);\
+    SREG _R,offset(_BR)					;\
+  CHK_OFFSET(_BR,SIGALIGN,1);\
+    SREG _F,offset(_BR)			;\
+    .set offset,offset+(SIGALIGN)
   
+// for updating signatures when 'rd' is a paired register (64-bit) in Zpsfoperand extension in RV32.
+#define RVTEST_SIGUPD_P64(_BR,_R,_R_HI,...)		 \
+ .if NARG(__VA_ARGS__) == 0				;\
+	RVTEST_SIGUPD_FID(_BR,_R,_R_HI)			;\
+ .else							;\
+	RVTEST_SIGUPD_FID(_BR,_R,_R_HI,_ARG1(__VA_OPT__(__VA_ARGS__,0)));\
+ .endif
+
+// for reading vxsat.OV flag in P-ext; and only reads the flag when Zicsr extension is present
+#ifdef pext_check_vxsat_ov
+#define RDOV(_F)\
+   rdov _F
+#else
+#define RDOV(_F)\
+   nop
+#endif
+
+// for updating signatures that include flagreg when 'rd' is a paired register (64-bit) in Zpsfoperand extension in RV32.
+#define RVTEST_SIGUPD_PK64(_BR,_R,_R_HI,_F,...)\
+  .if NARG(__VA_ARGS__) == 1                            ;\
+     .set offset,_ARG1(__VA_OPT__(__VA_ARGS__,0))	;\
+  .endif                                                ;\
+  CHK_OFFSET(_BR,REGWIDTH,0);\
+    SREG _R,offset(_BR)					;\
+  CHK_OFFSET(_BR,REGWIDTH,1);\
+    SREG _R_HI,offset+REGWIDTH(_BR)			;\
+    RDOV(_F)                                            ;\
+  CHK_OFFSET(_BR,REGWIDTH,1);\
+    SREG _F,offset+2*REGWIDTH(_BR)			;\
+    .set offset,offset+(3*REGWIDTH)
+
+// for updating signatures that include flagreg for P-ext saturation instructions (RV32/RV64).
+#define RVTEST_SIGUPD_PK(_BR,_R,_F,OFFSET)\
+  RVTEST_SIGUPD_FID(_BR,_R,_F,OFFSET)
+
 #define RVTEST_VALBASEUPD(_BR,...)\
   .if NARG(__VA_ARGS__) == 0;\
       addi _BR,_BR,2040;\
@@ -675,12 +779,12 @@ rvtest_data_end:
     jalr x0,0(tempreg)                       ;\
 6:  LA(tempreg, 4f                          ) ;\
     jalr x0,0(tempreg)                        ;\
-1:  .if adj & 2 == 2                         ;\
+1:  .if (adj & 2 == 2) && (label == 1b)      ;\
     .fill 2,1,0x00                          ;\
     .endif                                    ;\
     xori rd,rd, 0x1                           ;\
     beq x0,x0,6b                               ;\
-    .if adj & 2 == 2                              ;\
+    .if (adj & 2 == 2) && (label == 1b)     ;\
     .fill 2,1,0x00                          ;\
     .endif                                    ;\
     .if (imm/2) - 2 >= 0                      ;\
@@ -688,7 +792,7 @@ rvtest_data_end:
     .else                                     ;\
         .set num,0                            ;\
     .endif                                    ;\
-     .if label == 3f                          ;\
+     .ifc label, 3f                          ;\
         .set num,0                            ;\
     .endif                                    ;\
     .rept num                                 ;\
@@ -707,19 +811,19 @@ rvtest_data_end:
     .else                                     ;\
         .set num,0                            ;\
     .endif                                    ;\
-    .if label == 1b                          ;\
+    .ifc label, 1b                          ;\
         .set num,0                            ;\
     .endif                                    ;\
     .rept num                                 ;\
     nop                                       ;\
     .endr                                     ;\
-3:  .if adj & 2 == 2                              ;\
+3:  .if (adj & 2 == 2) && (label == 3f)      ;\
     .fill 2,1,0x00                          ;\
     .endif                                    ;\
     xori rd,rd, 0x3                           ;\
     LA(tempreg, 4f                          ) ;\
     jalr x0,0(tempreg)                        ;\
-    .if adj&2 == 2                              ;\
+    .if (adj&2 == 2) && (label == 3f)       ;\
     .fill 2,1,0x00                     ;\
     .endif                                    ;\
 4: LA(tempreg, 5b                            ) ;\
@@ -747,7 +851,7 @@ rvtest_data_end:
     .else                                     ;\
         .set num,0                            ;\
     .endif                                    ;\
-     .if label == 3f                          ;\
+     .ifc label, 3f                          ;\
         .set num,0                            ;\
     .endif                                    ;\
     .rept num                                 ;\
@@ -762,7 +866,7 @@ rvtest_data_end:
     .else                                     ;\
         .set num,0                            ;\
     .endif                                    ;\
-     .if label == 1b                          ;\
+     .ifc label, 1b                          ;\
         .set num,0                            ;\
     .endif                                    ;\
     .rept num                                 ;\
@@ -798,25 +902,25 @@ nop                                                                         ;\
 RVTEST_SIGUPD(swreg,destreg,offset) 
 //SREG destreg, offset(swreg);
 
-#define TEST_STORE_F(swreg,testreg,index,rs1,rs2,rs2_val,imm_val,offset,inst,adj,flagreg)   ;\
-LI(flagreg,rs2_val)                                                           ;\
-fmv.w.x rs2, flagreg                                                          ;\
+#define TEST_STORE_F(swreg,testreg,fcsr_val,rs1,rs2,imm_val,offset,inst,adj,flagreg,valaddr_reg, val_offset);\
+LOAD_MEM_VAL(FLREG, valaddr_reg, rs2, val_offset, testreg); \
 addi rs1,swreg,offset+adj                                                     ;\
 LI(testreg,imm_val)                                                         ;\
 sub rs1,rs1,testreg                                                          ;\
 inst rs2, imm_val(rs1)                                                      ;\
 nop                                                                         ;\
 nop                                                                         ;\
-csrrs flagreg, fflags, x0                                                   ;\
-RVTEST_SIGUPD(swreg,flagreg,offset)
+csrr flagreg, fcsr                                                         ;\
+RVTEST_SIGUPD(swreg,flagreg,offset+SIGALIGN)
 
-#define TEST_LOAD_F(swreg,testreg,index,rs1,destreg,imm_val,offset,inst,adj,flagreg)   ;\
-LA(rs1,rvtest_data+(index*4)+adj-imm_val)                                      ;\
-inst destreg, imm_val(rs1)                                                   ;\
-nop                                                                         ;\
-nop                                                                         ;\
-csrrs flagreg, fflags, x0                                                   ;\
-RVTEST_SIGUPD_F(swreg,destreg,flagreg,offset) 
+#define TEST_LOAD_F(swreg,testreg,fcsr_val,rs1,destreg,imm_val,inst,adj,flagreg)    ;\
+LA(rs1,rvtest_data+adj-imm_val)                                               ;\
+LI(testreg, fcsr_val); csrw fcsr, testreg                                                ;\
+inst destreg, imm_val(rs1)                                                              ;\
+nop                                                                                     ;\
+nop                                                                                     ;\
+csrr flagreg, fcsr                                                                      ;\
+RVTEST_SIGUPD_F(swreg,destreg,flagreg) 
 
 #define TEST_CSR_FIELD(ADDRESS,TEMP_REG,MASK_REG,NEG_MASK_REG,VAL,DEST_REG,OFFSET,BASE_REG) \
     LI(TEMP_REG,VAL);\
@@ -834,14 +938,14 @@ RVTEST_SIGUPD_F(swreg,destreg,flagreg,offset)
     RVTEST_SIGUPD(swreg,destreg,offset); \
     RVMODEL_IO_ASSERT_GPR_EQ(testreg, destreg, correctval)
 
-#define TEST_CASE_F(testreg, destreg, correctval, swreg, flagreg, offset, code... ) \
+#define TEST_CASE_F(testreg, destreg, correctval, swreg, flagreg, code... ) \
     code; \
-    RVTEST_SIGUPD_F(swreg,destreg,flagreg,offset); \
+    RVTEST_SIGUPD_F(swreg,destreg,flagreg); \
     RVMODEL_IO_ASSERT_GPR_EQ(testreg, destreg, correctval)
     
-#define TEST_CASE_FID(testreg, destreg, correctval, swreg, flagreg, offset, code... ) \
+#define TEST_CASE_FID(testreg, destreg, correctval, swreg, flagreg, code... ) \
     code; \
-    RVTEST_SIGUPD_FID(swreg,destreg,flagreg,offset); \
+    RVTEST_SIGUPD_FID(swreg,destreg,flagreg); \
     RVMODEL_IO_ASSERT_GPR_EQ(testreg, destreg, correctval)
 
 #define TEST_AUIPC(inst, destreg, correctval, imm, swreg, offset, testreg) \
@@ -852,47 +956,78 @@ RVTEST_SIGUPD_F(swreg,destreg,flagreg,offset)
       sub destreg, destreg, testreg; \
       )
 
-//Tests for a instructions with register-immediate operand
+//Tests for instructions with register-immediate operand
 #define TEST_IMM_OP( inst, destreg, reg, correctval, val, imm, swreg, offset, testreg) \
     TEST_CASE(testreg, destreg, correctval, swreg, offset, \
       LI(reg, MASK_XLEN(val)); \
       inst destreg, reg, SEXT_IMM(imm); \
     )
-    
+
 //Tests for floating-point instructions with a single register operand
-#define TEST_FPSR_OP( inst, destreg, freg, rm, correctval, valaddr_reg, val_offset, flagreg, swreg, offset, testreg) \
-    TEST_CASE_F(testreg, destreg, correctval, swreg, flagreg, offset, \
-      FLREG freg, val_offset(valaddr_reg); \
-      csrrwi x0, frm, rm; \
+#define TEST_FPSR_OP( inst, destreg, freg, rm, fcsr_val, correctval, valaddr_reg, val_offset, flagreg, swreg, testreg) \
+    TEST_CASE_F(testreg, destreg, correctval, swreg, flagreg, \
+      LOAD_MEM_VAL(FLREG, valaddr_reg, freg, val_offset, testreg); \
+      LI(testreg, fcsr_val); csrw fcsr, testreg; \
+      inst destreg, freg, rm; \
+      csrr flagreg, fcsr      ; \
+    )
+
+//Tests for floating-point instructions with a single register operand
+//This variant does not take the rm field and set it while writing the instruction
+#define TEST_FPSR_OP_NRM( inst, destreg, freg, fcsr_val, correctval, valaddr_reg, val_offset, flagreg, swreg, testreg) \
+    TEST_CASE_F(testreg, destreg, correctval, swreg, flagreg, \
+      LOAD_MEM_VAL(FLREG, valaddr_reg, freg, val_offset, testreg); \
+      LI(testreg, fcsr_val); csrw fcsr, testreg; \
       inst destreg, freg; \
-      csrrs flagreg, fflags, x0; \
+      csrr flagreg, fcsr      ; \
     )
     
 //Tests for floating-point instructions with a single register operand and integer destination register
-#define TEST_FPID_OP( inst, destreg, freg, rm, correctval, valaddr_reg, val_offset, flagreg, swreg, offset, testreg) \
-    TEST_CASE_FID(testreg, destreg, correctval, swreg, flagreg, offset, \
-      FLREG freg, val_offset(valaddr_reg); \
-      csrrwi x0, frm, rm; \
-      inst destreg, freg; \
-      csrrs flagreg, fflags, x0; \
-    )
+#define TEST_FPID_OP( inst, destreg, freg, rm, fcsr_val, correctval, valaddr_reg, val_offset, flagreg, swreg, testreg,load_instr) \
+    TEST_CASE_FID(testreg, destreg, correctval, swreg, flagreg, \
+      LOAD_MEM_VAL(load_instr, valaddr_reg, freg, val_offset, testreg); \
+      LI(testreg, fcsr_val); csrw fcsr, testreg; \
+      inst destreg, freg, rm; \
+      csrr flagreg, fcsr      ; \
+      )
     
 //Tests for floating-point instructions with a single register operand and integer operand register
-#define TEST_FPIO_OP( inst, destreg, freg, rm, correctval, valaddr_reg, val_offset, flagreg, swreg, offset, testreg) \
-    TEST_CASE_F(testreg, destreg, correctval, swreg, flagreg, offset, \
-      LREG freg, val_offset(valaddr_reg); \
-      csrrwi x0, frm, rm; \
-      inst destreg, freg; \
-      csrrs flagreg, fflags, x0; \
+#define TEST_FPIO_OP( inst, destreg, freg, rm, fcsr_val, correctval, valaddr_reg, val_offset, flagreg, swreg, testreg, load_instr) \
+    TEST_CASE_F(testreg, destreg, correctval, swreg, flagreg, \
+      LOAD_MEM_VAL(load_instr, valaddr_reg, freg, val_offset, testreg); \
+      LI(testreg, fcsr_val); csrw fcsr, testreg; \
+      inst destreg, freg, rm; \
+      csrr flagreg, fcsr; \
     )
 
+//Tests for floating-point instructions with a single register operand and integer destination register
+//This variant does not take the rm field and set it while writing the instruction
+#define TEST_FPID_OP_NRM( inst, destreg, freg, fcsr_val, correctval, valaddr_reg, val_offset, flagreg, swreg, testreg) \
+    TEST_CASE_FID(testreg, destreg, correctval, swreg, flagreg, \
+      LOAD_MEM_VAL(FLREG, valaddr_reg, freg, val_offset, testreg); \
+      LI(testreg, fcsr_val); csrw fcsr, testreg; \
+      inst destreg, freg; \
+      csrr flagreg, fcsr      ; \
+      )
+    
+//Tests for floating-point instructions with a single register operand and integer operand register
+//This variant does not take the rm field and set it while writing the instruction
+#define TEST_FPIO_OP_NRM( inst, destreg, freg, fcsr_val, correctval, valaddr_reg, val_offset, flagreg, swreg, testreg, load_instr) \
+    TEST_CASE_F(testreg, destreg, correctval, swreg, flagreg, \
+      LOAD_MEM_VAL(load_instr, valaddr_reg, freg, val_offset, testreg); \
+      LI(testreg, fcsr_val); csrw fcsr, testreg; \
+      inst destreg, freg; \
+      csrr flagreg, fcsr; \
+    )
+
+//Tests for instructions with register-register-immediate operands
 #define TEST_RRI_OP(inst, destreg, reg1, reg2, imm, correctval, val1, val2, swreg, offset, testreg) \
     TEST_CASE(testreg, destreg, correctval, swreg, offset, \
       LI(reg1, MASK_XLEN(val1)); \
       LI(reg2, MASK_XLEN(val2)); \
       inst destreg, reg1, reg2, imm; \
     )
-    
+
 //Tests for a instructions with register-register operand
 #define TEST_RI_OP(inst, destreg, reg2, imm, correctval, val1, val2, swreg, offset, testreg) \
     TEST_CASE(testreg, destreg, correctval, swreg, offset, \
@@ -909,33 +1044,45 @@ RVTEST_SIGUPD_F(swreg,destreg,flagreg,offset)
       inst destreg, reg1, reg2; \
     )
 //Tests for floating-point instructions with register-register operand
-#define TEST_FPRR_OP(inst, destreg, freg1, freg2, rm, correctval, valaddr_reg, val_offset, flagreg, swreg, offset, testreg) \
-    TEST_CASE_F(testreg, destreg, correctval, swreg, flagreg, offset, \
-      FLREG freg1, val_offset(valaddr_reg); \
-      FLREG freg2, val_offset+FREGWIDTH(valaddr_reg); \
-      csrrwi x0, frm, rm; \
+//This variant does not take the rm field and set it while writing the instruction
+#define TEST_FPRR_OP_NRM(inst, destreg, freg1, freg2, fcsr_val, correctval, valaddr_reg, val_offset, flagreg, swreg, testreg) \
+    TEST_CASE_F(testreg, destreg, correctval, swreg, flagreg, \
+      LOAD_MEM_VAL(FLREG, valaddr_reg, freg1, val_offset, testreg); \
+      LOAD_MEM_VAL(FLREG, valaddr_reg, freg2, (val_offset+FREGWIDTH), testreg); \
+      LI(testreg, fcsr_val); csrw fcsr, testreg; \
       inst destreg, freg1, freg2; \
-      csrrs flagreg, fflags, x0; \
+      csrr flagreg, fcsr; \
+    )
+
+//Tests for floating-point instructions with register-register operand
+#define TEST_FPRR_OP(inst, destreg, freg1, freg2, rm, fcsr_val, correctval, valaddr_reg, val_offset, flagreg, swreg, testreg) \
+    TEST_CASE_F(testreg, destreg, correctval, swreg, flagreg, \
+      LOAD_MEM_VAL(FLREG, valaddr_reg, freg1, val_offset, testreg); \
+      LOAD_MEM_VAL(FLREG, valaddr_reg, freg2, (val_offset+FREGWIDTH), testreg); \
+      LI(testreg, fcsr_val); csrw fcsr, testreg; \
+      inst destreg, freg1, freg2, rm; \
+      csrr flagreg, fcsr; \
     )
     
 //Tests for floating-point CMP instructions with register-register operand
-#define TEST_FCMP_OP(inst, destreg, freg1, freg2, correctval, valaddr_reg, val_offset, flagreg, swreg, offset, testreg) \
-    TEST_CASE_FID(testreg, destreg, correctval, swreg, flagreg, offset, \
-      FLREG freg1, val_offset(valaddr_reg); \
-      FLREG freg2, val_offset+FREGWIDTH(valaddr_reg); \
+#define TEST_FCMP_OP(inst, destreg, freg1, freg2, fcsr_val, correctval, valaddr_reg, val_offset, flagreg, swreg, testreg) \
+    TEST_CASE_FID(testreg, destreg, correctval, swreg, flagreg, \
+      LOAD_MEM_VAL(FLREG, valaddr_reg, freg1, val_offset, testreg); \
+      LOAD_MEM_VAL(FLREG, valaddr_reg, freg2, (val_offset+FREGWIDTH), testreg); \
+      LI(testreg, fcsr_val); csrw fcsr, testreg; \
       inst destreg, freg1, freg2; \
-      csrrs flagreg, fflags, x0; \
+      csrr flagreg, fcsr      ; \
     )
 
 //Tests for floating-point R4 type instructions
-#define TEST_FPR4_OP(inst, destreg, freg1, freg2, freg3, rm, correctval, valaddr_reg, val_offset, flagreg, swreg, offset, testreg) \
-    TEST_CASE_F(testreg, destreg, correctval, swreg, flagreg, offset, \
-      FLREG freg1, val_offset(valaddr_reg); \
-      FLREG freg2, val_offset+FREGWIDTH(valaddr_reg); \
-      FLREG freg3, val_offset+2*FREGWIDTH(valaddr_reg); \
-      csrrwi x0, frm, rm; \
-      inst destreg, freg1, freg2, freg3; \
-      csrrs flagreg, fflags, x0; \
+#define TEST_FPR4_OP(inst, destreg, freg1, freg2, freg3, rm , fcsr_val, correctval, valaddr_reg, val_offset, flagreg, swreg, testreg) \
+    TEST_CASE_F(testreg, destreg, correctval, swreg, flagreg, \
+      LOAD_MEM_VAL(FLREG, valaddr_reg, freg1, val_offset, testreg); \
+      LOAD_MEM_VAL(FLREG, valaddr_reg, freg2, (val_offset+FREGWIDTH), testreg); \
+      LOAD_MEM_VAL(FLREG, valaddr_reg, freg3, (val_offset+2*FREGWIDTH), testreg); \
+      LI(testreg, fcsr_val); csrw fcsr, testreg; \
+      inst destreg, freg1, freg2, freg3, rm; \
+      csrr flagreg, fcsr      ; \
     )
 
 #define TEST_CNOP_OP( inst, testreg, imm_val, swreg, offset) \
@@ -943,6 +1090,137 @@ RVTEST_SIGUPD_F(swreg,destreg,flagreg,offset)
       inst imm_val; \
       )
 
+//Tests for instructions with register-immediate operand and update the saturation flag
+#define TEST_PKIMM_OP( inst, destreg, reg, correctval, val, imm, flagreg, swreg, offset, testreg) \
+    TEST_CASE(testreg, destreg, correctval, swreg, offset, \
+      LI(reg, MASK_XLEN(val)); \
+      inst destreg, reg, SEXT_IMM(imm); \
+      rdov flagreg; \
+    )
+
+//Tests for instructions with register-register operand and update the saturation flag
+#define TEST_PKRR_OP(inst, destreg, reg1, reg2, correctval, val1, val2, flagreg, swreg, offset, testreg) \
+    LI(reg1, MASK_XLEN(val1)); \
+    LI(reg2, MASK_XLEN(val2)); \
+    inst destreg, reg1, reg2; \
+    rdov flagreg; \
+    RVTEST_SIGUPD_PK(swreg, destreg, flagreg, offset); \
+    RVMODEL_IO_ASSERT_GPR_EQ(testreg, destreg, correctval)
+
+//Tests for instructions with a single register operand and update the saturation flag
+#define TEST_PKR_OP( inst, destreg, reg, correctval, val, flagreg, swreg, offset, testreg) \
+    TEST_CASE_FID(testreg, destreg, correctval, swreg, flagreg, offset, \
+      LI(reg, MASK_XLEN(val)); \
+      inst destreg, reg; \
+      rdov flagreg; \
+    )
+
+#if __riscv_xlen == 32
+//Tests for a instruction with register pair operands for all its three operands
+#define TEST_P64_PPP_OP_32(inst, destreg, destreg_hi, reg1, reg1_hi, reg2, reg2_hi, correctval, correctval_hi, val1, val1_hi, val2, val2_hi, swreg, offset, testreg) \
+      LI(reg1, MASK_XLEN(val1)); \
+      LI(reg1_hi, MASK_XLEN(val1_hi)); \
+      LI(reg2, MASK_XLEN(val2)); \
+      LI(reg2_hi, MASK_XLEN(val2_hi)); \
+      inst destreg, reg1, reg2; \
+      RVTEST_SIGUPD_P64(swreg,destreg, destreg_hi, offset); \
+      RVMODEL_IO_ASSERT_GPR_EQ(testreg, destreg, correctval); \
+      RVMODEL_IO_ASSERT_GPR_EQ(testreg, destreg_hi, correctval_hi)
+
+#define TEST_PK64_PPP_OP_32(inst, destreg, destreg_hi, reg1, reg1_hi, reg2, reg2_hi, correctval, correctval_hi, val1, val1_hi, val2, val2_hi, flagreg, swreg, offset, testreg) \
+      LI(reg1, MASK_XLEN(val1)); \
+      LI(reg1_hi, MASK_XLEN(val1_hi)); \
+      LI(reg2, MASK_XLEN(val2)); \
+      LI(reg2_hi, MASK_XLEN(val2_hi)); \
+      inst destreg, reg1, reg2; \
+      RVTEST_SIGUPD_PK64(swreg,destreg, destreg_hi, flagreg, offset); \
+      RVMODEL_IO_ASSERT_GPR_EQ(testreg, destreg, correctval); \
+      RVMODEL_IO_ASSERT_GPR_EQ(testreg, destreg_hi, correctval_hi)
+
+#define TEST_P64_PPN_OP_32(inst, destreg, destreg_hi, reg1, reg1_hi, reg2, correctval, correctval_hi, val1, val1_hi, val2, swreg, offset, testreg) \
+      LI(reg1, MASK_XLEN(val1)); \
+      LI(reg1_hi, MASK_XLEN(val1_hi)); \
+      LI(reg2, MASK_XLEN(val2)); \
+      inst destreg, reg1, reg2; \
+      RVTEST_SIGUPD_P64(swreg, destreg, destreg_hi, offset); \
+      RVMODEL_IO_ASSERT_GPR_EQ(testreg, destreg, correctval); \
+      RVMODEL_IO_ASSERT_GPR_EQ(testreg, destreg_hi, correctval_hi)
+
+#define TEST_P64_PNN_OP_32(inst, destreg, destreg_hi, reg1, reg2, correctval, correctval_hi, val1, val2, swreg, offset, testreg) \
+      LI(reg1, MASK_XLEN(val1)); \
+      LI(reg2, MASK_XLEN(val2)); \
+      inst destreg, reg1, reg2; \
+      RVTEST_SIGUPD_P64(swreg, destreg, destreg_hi, offset); \
+      RVMODEL_IO_ASSERT_GPR_EQ(testreg, destreg, correctval); \
+      RVMODEL_IO_ASSERT_GPR_EQ(testreg, destreg_hi, correctval_hi)
+
+#define TEST_PK64_PNN_OP_32(inst, destreg, destreg_hi, reg1, reg2, correctval, correctval_hi, val1, val2, flagreg, swreg, offset, testreg) \
+      LI(reg1, MASK_XLEN(val1)); \
+      LI(reg2, MASK_XLEN(val2)); \
+      inst destreg, reg1, reg2; \
+      RVTEST_SIGUPD_PK64(swreg, destreg, destreg_hi, flagreg, offset); \
+      RVMODEL_IO_ASSERT_GPR_EQ(testreg, destreg, correctval); \
+      RVMODEL_IO_ASSERT_GPR_EQ(testreg, destreg_hi, correctval_hi)
+
+#define TEST_P64_NPN_OP_32(inst, destreg, reg1, reg1_hi, reg2, correctval, val1, val1_hi, val2, swreg, offset, testreg) \
+      LI(reg1, MASK_XLEN(val1)); \
+      LI(reg1_hi, MASK_XLEN(val1_hi)); \
+      LI(reg2, MASK_XLEN(val2)); \
+      inst destreg, reg1, reg2; \
+      RVTEST_SIGUPD(swreg,destreg,offset); \
+      RVMODEL_IO_ASSERT_GPR_EQ(testreg, destreg, correctval);
+
+#define TEST_P64_NP_OP_32(inst, destreg, reg1, reg1_hi, correctval, val1, val1_hi, imm_val, swreg, offset, testreg) \
+      LI(reg1, MASK_XLEN(val1)); \
+      LI(reg1_hi, MASK_XLEN(val1_hi)); \
+      inst destreg, reg1, imm_val; \
+      RVTEST_SIGUPD(swreg,destreg,offset); \
+      RVMODEL_IO_ASSERT_GPR_EQ(testreg, destreg, correctval);
+
+//Tests for a instruction with pair register rd, pair register rs1 and pair register rs2
+#define TEST_P64_PPP_OP(inst, rd, rd_hi, rs1, rs1_hi, rs2, rs2_hi, correctval, correctval_hi, rs1_val, rs1_val_hi, rs2_val, rs2_val_hi, swreg, offset, testreg) \
+    TEST_P64_PPP_OP_32(inst, rd, rd_hi, rs1, rs1_hi, rs2, rs2_hi, correctval, correctval_hi, rs1_val, rs1_val_hi, rs2_val, rs2_val_hi, swreg, offset, testreg)
+#define TEST_PK64_PPP_OP(inst, rd, rd_hi, rs1, rs1_hi, rs2, rs2_hi, correctval, correctval_hi, rs1_val, rs1_val_hi, rs2_val, rs2_val_hi, flagreg, swreg, offset, testreg) \
+    TEST_PK64_PPP_OP_32(inst, rd, rd_hi, rs1, rs1_hi, rs2, rs2_hi, correctval, correctval_hi, rs1_val, rs1_val_hi, rs2_val, rs2_val_hi, flagreg, swreg, offset, testreg)
+//Tests for a instruction with pair register rd, pair register rs1 and normal register rs2
+#define TEST_P64_PPN_OP(inst, rd, rd_hi, rs1, rs1_hi, rs2, correctval, correctval_hi, rs1_val, rs1_val_hi, rs2_val, swreg, offset, testreg) \
+    TEST_P64_PPN_OP_32(inst, rd, rd_hi, rs1, rs1_hi, rs2, correctval, correctval_hi, rs1_val, rs1_val_hi, rs2_val, swreg, offset, testreg)
+//Tests for a instruction with pair register rd, normal register rs1 and normal register rs2
+#define TEST_P64_PNN_OP(inst, rd, rd_hi, rs1, rs2, correctval, correctval_hi, rs1_val, rs2_val, swreg, offset, testreg) \
+    TEST_P64_PNN_OP_32(inst, rd, rd_hi, rs1, rs2, correctval, correctval_hi, rs1_val, rs2_val, swreg, offset, testreg)
+//Tests for a instruction with pair register rd, normal register rs1 and normal register rs2
+#define TEST_PK64_PNN_OP(inst, rd, rd_hi, rs1, rs2, correctval, correctval_hi, rs1_val, rs2_val, flagreg, swreg, offset, testreg) \
+    TEST_PK64_PNN_OP_32(inst, rd, rd_hi, rs1, rs2, correctval, correctval_hi, rs1_val, rs2_val, flagreg, swreg, offset, testreg)
+//Tests for a instruction with normal register rd, pair register rs1 and normal register rs2
+#define TEST_P64_NPN_OP(inst, rd, rs1, rs1_hi, rs2, correctval, correctval_hi, rs1_val, rs1_val_hi, rs2_val, swreg, offset, testreg) \
+    TEST_P64_NPN_OP_32(inst, rd, rs1, rs1_hi, rs2, correctval, correctval_hi, rs1_val, rs1_val_hi, rs2_val, swreg, offset, testreg)
+//Tests for a instruction with normal register rd, pair register rs1
+#define TEST_P64_NP_OP(inst, rd, rs1, rs1_hi, correctval, correctval_hi, rs1_val, rs1_val_hi, imm_val, swreg, offset, testreg) \
+    TEST_P64_NP_OP_32(inst, rd, rs1, rs1_hi, correctval, correctval_hi, rs1_val, rs1_val_hi, imm_val, swreg, offset, testreg)
+
+#else
+
+// When in rv64, there are no instructions with pair operand, so Macro is redefined to normal TEST_RR_OP
+#define TEST_P64_PPP_OP(inst, rd, rd_hi, rs1, rs1_hi, rs2, rs2_hi, correctval, correctval_hi, rs1_val, rs1_val_hi, rs2_val, rs2_val_hi, swreg, offset, testreg) \
+    TEST_RR_OP(inst, rd, rs1, rs2, correctval, rs1_val, rs2_val, swreg, offset, testreg)
+#define TEST_PK64_PPP_OP(inst, rd, rd_hi, rs1, rs1_hi, rs2, rs2_hi, correctval, correctval_hi, rs1_val, rs1_val_hi, rs2_val, rs2_val_hi, flagreg, swreg, offset, testreg) \
+    TEST_PKRR_OP(inst, rd, rs1, rs2, correctval, rs1_val, rs2_val, flagreg, swreg, offset, testreg)
+#define TEST_P64_PPN_OP(inst, rd, rd_hi, rs1, rs1_hi, rs2, correctval, correctval_hi, rs1_val, rs1_val_hi, rs2_val, swreg, offset, testreg) \
+    TEST_RR_OP(inst, rd, rs1, rs2, correctval, rs1_val, rs2_val, swreg, offset, testreg)
+#define TEST_P64_PNN_OP(inst, rd, rd_hi, rs1, rs2, correctval, correctval_hi, rs1_val, rs2_val, swreg, offset, testreg) \
+    TEST_RR_OP(inst, rd, rs1, rs2, correctval, rs1_val, rs2_val, swreg, offset, testreg)
+#define TEST_PK64_PNN_OP(inst, rd, rd_hi, rs1, rs2, correctval, correctval_hi, rs1_val, rs2_val, flagreg, swreg, offset, testreg) \
+    TEST_PKRR_OP(inst, rd, rs1, rs2, correctval, rs1_val, rs2_val, flagreg, swreg, offset, testreg)
+#define TEST_P64_NPN_OP(inst, rd, rs1, rs1_hi, rs2, correctval, correctval_hi, rs1_val, rs1_val_hi, rs2_val, swreg, offset, testreg) \
+    TEST_RR_OP(inst, rd, rs1, rs2, correctval, rs1_val, rs2_val, swreg, offset, testreg)
+#define TEST_P64_NP_OP(inst, rd, rs1, rs1_hi, correctval, correctval_hi, rs1_val, rs1_val_hi, imm_val, swreg, offset, testreg) \
+    TEST_IMM_OP(inst, rd, rs1, correctval, rs1_val, imm_val, swreg, offset, testreg)
+
+#endif
+
+
+
+
 #define TEST_CMV_OP( inst, destreg, reg, correctval, val2, swreg, offset, testreg) \
     TEST_CASE(testreg, destreg, correctval, swreg, offset, \
       LI(reg, MASK_XLEN(val2)); \
@@ -968,6 +1246,10 @@ RVTEST_SIGUPD_F(swreg,destreg,flagreg,offset)
       inst destreg, x2,imm; \
       )
 
+//Tests for instructions with a single register operand
+#define TEST_RD_OP(inst, destreg, reg1, correctval, val1, swreg, offset, testreg) \
+  TEST_CMV_OP(inst, destreg, reg1, correctval, val1, swreg, offset, testreg)
+
 #define TEST_CBRANCH_OP(inst, tempreg, reg2, val2, imm, label, swreg, offset) \
     LI(reg2, MASK_XLEN(val2))                  ;\
     j 2f                                      ;\
@@ -982,7 +1264,7 @@ RVTEST_SIGUPD_F(swreg,destreg,flagreg,offset)
     .else                                     ;\
         .set num,0                            ;\
     .endif                                    ;\
-    .if label == 3f                           ;\
+    .ifc label, 3f                           ;\
         .set num,0                            ;\
     .endif                                    ;\
     .rept num                                 ;\
@@ -999,7 +1281,7 @@ RVTEST_SIGUPD_F(swreg,destreg,flagreg,offset)
     .else                                     ;\
         .set num,0                            ;\
     .endif                                    ;\
-     .if label == 1b                          ;\
+     .ifc label, 1b                          ;\
         .set num,0                            ;\
     .endif                                    ;\
     .rept num                                 ;\
@@ -1025,7 +1307,7 @@ RVTEST_SIGUPD_F(swreg,destreg,flagreg,offset)
     .else                                     ;\
         .set num,0                            ;\
     .endif                                    ;\
-    .if label == 3f                           ;\
+    .ifc label, 3f                           ;\
         .set num,0                            ;\
     .endif                                    ;\
     .rept num                                 ;\
@@ -1042,7 +1324,7 @@ RVTEST_SIGUPD_F(swreg,destreg,flagreg,offset)
     .else                                     ;\
         .set num,0                            ;\
     .endif                                    ;\
-     .if label == 1b                          ;\
+     .ifc label, 1b                          ;\
         .set num,0                            ;\
     .endif                                    ;\
     .rept num                                 ;\
@@ -1068,7 +1350,7 @@ RVTEST_SIGUPD_F(swreg,destreg,flagreg,offset)
     .else                                     ;\
         .set num,0                            ;\
     .endif                                    ;\
-    .if label == 3f                           ;\
+    .ifc label, 3f                           ;\
         .set num,0                            ;\
     .endif                                    ;\
     .rept num                                 ;\
@@ -1085,7 +1367,7 @@ RVTEST_SIGUPD_F(swreg,destreg,flagreg,offset)
     .else                                     ;\
         .set num,0                            ;\
     .endif                                    ;\
-     .if label == 1b                          ;\
+     .ifc label, 1b                          ;\
         .set num,0                            ;\
     .endif                                    ;\
     .rept num                                 ;\
